from git_autograder.answers.answers_record import GitAutograderAnswersRecord
from git_autograder.answers.rules.answer_rule import AnswerRule


class NotEmptyRule(AnswerRule):
    EMPTY = "Answer for {question} is empty."

    def apply(self, answer: GitAutograderAnswersRecord) -> None:
        if answer.answer.strip() != "":
            raise Exception(self.EMPTY.format(question=answer.question))



from typing import Iterator, List, Optional, Tuple, Union

from difflib_parser import DifflibParser
from git import Commit, Diff, DiffIndex
from git.diff import Lit_change_type

from git_autograder.commit import GitAutograderCommit
from git_autograder.diff.diff import GitAutograderDiff


class GitAutograderDiffHelper:
    def __init__(
        self,
        a: Union[Commit, GitAutograderCommit],
        b: Union[Commit, GitAutograderCommit],
    ) -> None:
        a_commit = self.__get_commit(a)
        b_commit = self.__get_commit(b)
        self.diff_index: DiffIndex[Diff] = a_commit.diff(b_commit)

    def __get_commit(self, commit: Union[Commit, GitAutograderCommit]) -> Commit:
        if isinstance(commit, Commit):
            return commit
        return commit.commit

    @staticmethod
    def get_file_diff(
        a: Union[Commit, GitAutograderCommit],
        b: Union[Commit, GitAutograderCommit],
        file_path: str,
    ) -> Optional[Tuple["GitAutograderDiff", Lit_change_type]]:
        """Returns file difference between two commits across ALL change types."""
        # Based on the expectation that there can only exist one change type per file in a diff
        diff_helper = GitAutograderDiffHelper(a, b)
        change_types: List[Lit_change_type] = ["A", "D", "R", "M", "T"]
        for change_type in change_types:
            for change in diff_helper.iter_changes(change_type):
                if change.diff_parser is None or change.edited_file_path != file_path:
                    continue
                return change, change_type
        return None

    def iter_changes(self, change_type: Lit_change_type) -> Iterator[GitAutograderDiff]:
        for change in self.diff_index.iter_change_type(change_type):
            original_file_rawpath = change.a_rawpath
            edited_file_rawpath = change.b_rawpath
            original_file_path = (
                original_file_rawpath.decode("utf-8")
                if original_file_rawpath is not None
                else None
            )
            edited_file_path = (
                edited_file_rawpath.decode("utf-8")
                if edited_file_rawpath is not None
                else None
            )
            original_file_blob = change.a_blob
            edited_file_blob = change.b_blob
            original_file = (
                original_file_blob.data_stream.read().decode("utf-8")
                if original_file_blob is not None
                else None
            )
            edited_file = (
                edited_file_blob.data_stream.read().decode("utf-8")
                if edited_file_blob is not None
                else None
            )

            diff_parser = (
                DifflibParser(original_file.split("\n"), edited_file.split("\n"))
                if original_file is not None and edited_file is not None
                else None
            )

            yield GitAutograderDiff(
                change_type=change_type,
                diff=change,
                original_file_path=original_file_path,
                edited_file_path=edited_file_path,
                original_file=original_file,
                edited_file=edited_file,
                diff_parser=diff_parser,
            )



import os
from contextlib import contextmanager
from typing import Iterator, Optional, Union, TextIO

from git import Repo


class FileHelper:
    def __init__(self, repo: Repo) -> None:
        self.repo = repo

    @contextmanager
    def file_or_none(
        self, path: Union[str, os.PathLike[str]]
    ) -> Iterator[Optional[TextIO]]:
        file_path = os.path.join(self.repo.working_dir, path)
        if not os.path.isfile(file_path):
            yield None
        else:
            with open(file_path, "r") as file:
                yield file

    @contextmanager
    def file(self, path: Union[str, os.PathLike[str]]) -> Iterator[TextIO]:
        file_path = os.path.join(self.repo.working_dir, path)
        with open(file_path, "r") as file:
            yield file



from typing import Optional, Union

from git import Repo
from git.types import Commit_ish

from git_autograder.commit import GitAutograderCommit


class CommitHelper:
    def __init__(self, repo: Repo) -> None:
        self.repo = repo

    def commit(self, rev: Optional[Union[Commit_ish, str]]) -> GitAutograderCommit:
        c = self.repo.commit(rev)
        return GitAutograderCommit(c)



__all__ = ["BranchHelper", "CommitHelper", "RemoteHelper", "FileHelper"]

from .branch_helper import BranchHelper
from .commit_helper import CommitHelper
from .remote_helper import RemoteHelper
from .file_helper import FileHelper



from typing import Optional

from git import Repo

from git_autograder.branch import GitAutograderBranch
from git_autograder.exception import GitAutograderInvalidStateException


class BranchHelper:
    MISSING_BRANCH = "Branch {branch} is missing."

    def __init__(self, repo: Repo) -> None:
        self.repo = repo

    def branch_or_none(self, branch_name: str) -> Optional[GitAutograderBranch]:
        for head in self.repo.heads:
            if head.name == branch_name:
                return GitAutograderBranch(head)
        return None

    def branch(self, branch_name: str) -> GitAutograderBranch:
        b = self.branch_or_none(branch_name)
        if b is None:
            raise GitAutograderInvalidStateException(
                self.MISSING_BRANCH.format(branch=branch_name)
            )
        return b

    def has_branch(self, branch_name: str) -> bool:
        return self.branch_or_none(branch_name) is not None



from typing import Any, List, Optional, Sequence, Union

from git import Commit, Stats


class GitAutograderCommit:
    def __init__(self, commit: Commit) -> None:
        self.commit = commit

    def __eq__(self, value: Any) -> bool:
        if not isinstance(value, GitAutograderCommit):
            return False
        return value.commit == self.commit

    @property
    def hexsha(self) -> str:
        return self.commit.hexsha

    @property
    def stats(self) -> Stats:
        return self.commit.stats

    @property
    def parents(self) -> Sequence["GitAutograderCommit"]:
        return [GitAutograderCommit(parent) for parent in self.commit.parents]

    @property
    def branches(self) -> List[str]:
        """
        Returns the branches that contain the current commit.
        """
        containing_branches = self.commit.repo.git.branch("--contains", self.hexsha)
        return [line[2:] for line in containing_branches.split("\n")]

    def is_child(self, parent: Union[Commit, "GitAutograderCommit"]) -> bool:
        def _is_child(child: Commit, parent: Commit) -> bool:
            if child == parent:
                return True

            res = False
            for child_parent in child.parents:
                res |= _is_child(child_parent, parent)

            return res

        return _is_child(
            self.commit, parent if isinstance(parent, Commit) else parent.commit
        )

    def file_change_type(self, file_name: str) -> Optional[str]:
        if file_name not in self.stats.files:
            return None
        return self.stats.files[file_name]["change_type"]



from dataclasses import dataclass
from typing import List, Optional

from git_autograder.answers.answers_record import GitAutograderAnswersRecord
from git_autograder.answers.rules.answer_rule import AnswerRule
from git_autograder.exception import (
    GitAutograderInvalidStateException,
    GitAutograderWrongAnswerException,
)


@dataclass
class GitAutograderAnswers:
    MISSING_QUESTION = "Missing question {question} in answers file."

    questions: List[str]
    answers: List[str]

    @property
    def qna(self) -> List[GitAutograderAnswersRecord]:
        return list(
            map(
                lambda a: GitAutograderAnswersRecord.from_tuple(a),
                zip(self.questions, self.answers),
            )
        )

    def __getitem__(self, key: int) -> GitAutograderAnswersRecord:
        question = self.questions[key]
        answer = self.answers[key]
        return GitAutograderAnswersRecord.from_tuple((question, answer))

    def __len__(self) -> int:
        return len(self.questions)

    def question_or_none(self, question: str) -> Optional[GitAutograderAnswersRecord]:
        """
        Retrieves the record given a question.

        :returns: GitAutograderAnswersRecord if present, else None.
        :rtype: Optional[GitAutograderAnswersRecord]
        :raises GitAutograderInvalidStateException: if question is not present.
        """
        for i, q in enumerate(self.questions):
            if question == q:
                return GitAutograderAnswersRecord.from_tuple((q, self.answers[i]))
        return None

    def question(self, question: str) -> GitAutograderAnswersRecord:
        """
        Retrieves the record given a question.

        :returns: GitAutograderAnswersRecord if present.
        :rtype: GitAutograderAnswersRecord
        :raises GitAutograderInvalidStateException: if question is not present.
        """
        record = self.question_or_none(question)
        if record is None:
            raise GitAutograderInvalidStateException(
                self.MISSING_QUESTION.format(question=question)
            )
        return record

    def validate_question(
        self, question: str, rules: List[AnswerRule]
    ) -> "GitAutograderAnswers":
        """
        Validates that a given GitAutograderAnswersRecord passes a set of rules.

        :raises GitAutograderWrongAnswerException: when a rule is violated.
        """
        q = self.question(question)

        for rule in rules:
            try:
                rule.apply(q)
            except Exception as e:
                raise GitAutograderWrongAnswerException([str(e)])

        return self



import os
from datetime import datetime
from pathlib import Path
from typing import List, Optional

import pytz
from git import Repo

from git_autograder.answers.answers import GitAutograderAnswers
from git_autograder.answers.answers_parser import GitAutograderAnswersParser
from git_autograder.exception import (
    GitAutograderInvalidStateException,
    GitAutograderWrongAnswerException,
)
from git_autograder.helpers.branch_helper import BranchHelper
from git_autograder.helpers.commit_helper import CommitHelper
from git_autograder.helpers.file_helper import FileHelper
from git_autograder.helpers.remote_helper import RemoteHelper
from git_autograder.output import GitAutograderOutput
from git_autograder.status import GitAutograderStatus


class GitAutograderRepo:
    def __init__(
        self,
        exercise_name: str,
        repo_path: str | os.PathLike,
    ) -> None:
        # TODO: We should not be starting the grading at the point of initializing, but
        # we're keeping this because of the exception system
        self.started_at = self.__now()
        self.exercise_name = exercise_name
        self.repo_path = repo_path

        self.repo: Repo = Repo(self.repo_path)

        self.branches: BranchHelper = BranchHelper(self.repo)
        self.commits: CommitHelper = CommitHelper(self.repo)
        self.remotes: RemoteHelper = RemoteHelper(self.repo)
        self.files: FileHelper = FileHelper(self.repo)
        self.__answers_parser: Optional[GitAutograderAnswersParser] = None
        self.__answers: Optional[GitAutograderAnswers] = None

    @property
    def answers(self) -> GitAutograderAnswers:
        """Parses a QnA file (answers.txt). Verifies that the file exists."""
        # We need to use singleton patterns here since we want to avoid repeatedly parsing
        # These are all optional to start since the grader might not require answers
        if self.__answers_parser is None:
            answers_file_path = Path(self.repo_path) / "answers.txt"
            # Use singleton for answers parser
            try:
                self.__answers_parser = GitAutograderAnswersParser(answers_file_path)
            except Exception as e:
                raise GitAutograderInvalidStateException(
                    str(e),
                )

        if self.__answers is None:
            self.__answers = self.__answers_parser.answers

        return self.__answers

    @staticmethod
    def __now() -> datetime:
        return datetime.now(tz=pytz.UTC)

    def to_output(
        self, comments: List[str], status: Optional[GitAutograderStatus] = None
    ) -> GitAutograderOutput:
        """
        Creates a GitAutograderOutput object.

        If there is no status provided, the status will be inferred from the comments.
        """
        return GitAutograderOutput(
            exercise_name=self.exercise_name,
            started_at=self.started_at,
            completed_at=self.__now(),
            comments=comments,
            status=(
                GitAutograderStatus.SUCCESSFUL
                if len(comments) == 0
                else GitAutograderStatus.UNSUCCESSFUL
            )
            if status is None
            else status,
        )

    def wrong_answer(self, comments: List[str]) -> GitAutograderWrongAnswerException:
        return GitAutograderWrongAnswerException(comments)



from typing import Optional

from git import Repo

from git_autograder.exception import GitAutograderInvalidStateException
from git_autograder.remote import GitAutograderRemote


class RemoteHelper:
    MISSING_REMOTE = "Remote {remote} is missing."

    def __init__(self, repo: Repo) -> None:
        self.repo = repo

    def remote_or_none(self, remote_name: str) -> Optional[GitAutograderRemote]:
        for remote in self.repo.remotes:
            if remote.name == remote_name:
                return GitAutograderRemote(remote)
        return None

    def remote(self, remote_name: str) -> GitAutograderRemote:
        r = self.remote_or_none(remote_name)
        if r is None:
            raise GitAutograderInvalidStateException(
                self.MISSING_REMOTE.format(remote=remote_name)
            )
        return r

    def has_remote(self, remote_name: str) -> bool:
        return self.remote_or_none(remote_name) is not None



from typing import List

from git_autograder.answers.answers_record import GitAutograderAnswersRecord
from git_autograder.answers.rules.answer_rule import AnswerRule


class HasExactListRule(AnswerRule):
    INCORRECT_ORDERED = "Answer for {question} does not contian all of the right answers. Ensure that they follow the order specified."
    INCORRECT_UNORDERED = (
        "Answer for {question} does not contain all of the right answers."
    )

    def __init__(
        self, values: List[str], ordered: bool = False, is_case_sensitive: bool = False
    ) -> None:
        self.values = values
        self.ordered = ordered
        self.is_case_sensitive = is_case_sensitive

    def apply(self, answer: GitAutograderAnswersRecord) -> None:
        expected = (
            [v.lower() for v in self.values] if self.is_case_sensitive else self.values
        )
        given = (
            [v.lower() for v in answer.answer_as_list()]
            if self.is_case_sensitive
            else answer.answer_as_list()
        )
        if self.ordered and expected != given:
            raise Exception(self.INCORRECT_ORDERED.format(question=answer.question))
        elif set(expected).intersection(set(given)) != len(expected):
            raise Exception(self.INCORRECT_UNORDERED.format(question=answer.question))



from typing import Any, List

from git import Remote


class GitAutograderRemote:
    def __init__(self, remote: Remote) -> None:
        self.remote = remote

    def __eq__(self, value: Any) -> bool:
        if not isinstance(value, GitAutograderRemote):
            return False
        return value.remote == self.remote

    def track_branches(self, branches: List[str]) -> None:
        # We start with filtering main because it should be the default branch that
        # exists even on local machines.
        tracked = {"main"}
        for remote in self.remote.refs:
            for b in branches:
                if b not in tracked or f"{self.remote.name}/{b}" != remote.name:
                    continue
                tracked.add(b)
                self.remote.repo.git.checkout("-b", b, f"{self.remote.name}/{b}")
                break



from typing import Any, List

from git import Head

from git_autograder.commit import GitAutograderCommit
from git_autograder.diff import GitAutograderDiffHelper
from git_autograder.exception import GitAutograderInvalidStateException


class GitAutograderBranch:
    MISSING_START_COMMIT = "Branch {branch} is missing the Git Mastery start commit"
    MISSING_COMMITS = "Branch {branch} is missing any commits"

    def __init__(self, branch: Head) -> None:
        self.branch = branch

    def __eq__(self, value: Any) -> bool:
        if not isinstance(value, GitAutograderBranch):
            return False
        return value.branch == self.branch

    @property
    def name(self) -> str:
        return self.branch.name

    @property
    def commits(self) -> List[GitAutograderCommit]:
        """Retrieve the available commits of a given branch."""
        commits: List[GitAutograderCommit] = []
        for commit in self.branch.repo.iter_commits(self.branch):
            commits.append(GitAutograderCommit(commit))

        return commits

    @property
    def start_commit(self) -> GitAutograderCommit:
        """
        Find the Git Mastery start commit from the given branch.

        Raises exceptions if the branch has no commits or if the start tag is not
        present.
        """
        commits = self.commits

        if len(commits) == 0:
            raise GitAutograderInvalidStateException(
                self.MISSING_COMMITS.format(branch=self.name)
            )

        first_commit = commits[-1]
        first_commit_hash = first_commit.hexsha

        start_tag_name = f"git-mastery-start-{first_commit_hash[:7]}"

        start_tag = None
        for tag in self.branch.repo.tags:
            if str(tag) == start_tag_name:
                start_tag = tag
                break

        if start_tag is None:
            raise GitAutograderInvalidStateException(
                self.MISSING_START_COMMIT.format(branch=self.name)
            )

        return GitAutograderCommit(start_tag.commit)

    @property
    def user_commits(self) -> List[GitAutograderCommit]:
        """
        Retrieves only the user commits from a given branch.

        Raises exceptions if the branch has no commits or start tag is not present.
        """
        start_commit = self.start_commit
        commits = self.commits
        commits_asc = list(reversed(commits))
        start_commit_index = commits_asc.index(start_commit)
        user_commits = commits_asc[start_commit_index + 1 :]

        return user_commits

    @property
    def latest_commit(self) -> GitAutograderCommit:
        return self.user_commits[-1]

    def has_non_empty_commits(self) -> bool:
        """Returns if a given branch has any non-empty commits."""
        for commit in self.user_commits:
            if len(commit.stats.files) > 0:
                return True
        return False

    def has_edited_file(self, file_path: str) -> bool:
        """Returns if a given file has been edited in a given branch."""
        diff_helper = GitAutograderDiffHelper(self.start_commit, self.latest_commit)
        for diff in diff_helper.iter_changes("M"):
            if diff.edited_file_path == file_path:
                return True
        return False

    def has_added_file(self, file_path: str) -> bool:
        """Returns if a given file has been added in a given branch."""
        diff_helper = GitAutograderDiffHelper(self.start_commit, self.latest_commit)
        for diff in diff_helper.iter_changes("A"):
            if diff.edited_file_path == file_path:
                return True
        return False

    def checkout(self) -> None:
        self.branch.checkout()



import os
from contextlib import contextmanager
from datetime import datetime
from typing import Callable, Iterator, List, Optional
from unittest import mock

import pytz
from git import Repo
from repo_smith.initialize_repo import RepoInitializer, initialize_repo

from git_autograder.exception import (
    GitAutograderInvalidStateException,
    GitAutograderWrongAnswerException,
)
from git_autograder.output import GitAutograderOutput
from git_autograder.repo import GitAutograderRepo
from git_autograder.status import GitAutograderStatus


def attach_start_tag(repo_initializer: RepoInitializer, step_id: str) -> None:
    def hook(r: Repo) -> None:
        all_commits = list(r.iter_commits())
        first_commit = list(reversed(all_commits))[0]
        first_commit_hash = first_commit.hexsha[:7]
        start_tag = f"git-mastery-start-{first_commit_hash}"
        r.create_tag(start_tag)

    repo_initializer.add_post_hook(step_id, hook)


def set_env(**kwargs) -> mock._patch_dict:
    return mock.patch.dict(os.environ, kwargs, clear=True)


class GitAutograderTestLoader:
    def __init__(
        self,
        exercise_name: str,
        grade_func: Callable[[GitAutograderRepo], GitAutograderOutput],
    ) -> None:
        self.exercise_name = exercise_name
        self.grade_func = grade_func

    @contextmanager
    def load(
        self,
        spec_path: str,
        step_id: str,
        setup: Optional[Callable[[Repo], None]] = None,
    ) -> Iterator[GitAutograderOutput]:
        repo_initializer = initialize_repo(spec_path)
        attach_start_tag(repo_initializer, step_id)
        with repo_initializer.initialize() as r:
            if setup is not None:
                setup(r)
            output: Optional[GitAutograderOutput] = None
            started_at = datetime.now(tz=pytz.UTC)
            try:
                autograder = GitAutograderRepo(
                    exercise_name=self.exercise_name,
                    repo_path=r.working_dir,
                )
                output = self.grade_func(autograder)
            except (
                GitAutograderInvalidStateException,
                GitAutograderWrongAnswerException,
            ) as e:
                output = GitAutograderOutput(
                    exercise_name=self.exercise_name,
                    started_at=started_at,
                    completed_at=datetime.now(tz=pytz.UTC),
                    comments=[e.message] if isinstance(e.message, str) else e.message,
                    status=(
                        GitAutograderStatus.ERROR
                        if isinstance(e, GitAutograderInvalidStateException)
                        else GitAutograderStatus.UNSUCCESSFUL
                    ),
                )
            except Exception as e:
                # Unexpected exception
                output = GitAutograderOutput(
                    exercise_name=self.exercise_name,
                    started_at=None,
                    completed_at=None,
                    comments=[str(e)],
                    status=GitAutograderStatus.ERROR,
                )

            assert output is not None
            yield output


@contextmanager
def setup_autograder(
    exercise_name: str,
    spec_path: str,
    step_id: str,
    grade_func: Callable[[GitAutograderRepo], GitAutograderOutput],
    setup: Optional[Callable[[Repo], None]] = None,
) -> Iterator[GitAutograderOutput]:
    repo_initializer = initialize_repo(spec_path)
    attach_start_tag(repo_initializer, step_id)
    with repo_initializer.initialize() as r:
        if setup is not None:
            setup(r)
        output: Optional[GitAutograderOutput] = None
        started_at = datetime.now(tz=pytz.UTC)
        try:
            autograder = GitAutograderRepo(
                exercise_name=exercise_name, repo_path=r.working_dir
            )
            output = grade_func(autograder)
        except (
            GitAutograderInvalidStateException,
            GitAutograderWrongAnswerException,
        ) as e:
            output = GitAutograderOutput(
                exercise_name=exercise_name,
                started_at=started_at,
                completed_at=datetime.now(tz=pytz.UTC),
                comments=[e.message] if isinstance(e.message, str) else e.message,
                status=(
                    GitAutograderStatus.ERROR
                    if isinstance(e, GitAutograderInvalidStateException)
                    else GitAutograderStatus.UNSUCCESSFUL
                ),
            )
        except Exception as e:
            # Unexpected exception
            output = GitAutograderOutput(
                exercise_name=exercise_name,
                started_at=None,
                completed_at=None,
                comments=[str(e)],
                status=GitAutograderStatus.ERROR,
            )

        assert output is not None
        yield output


def assert_output(
    output: GitAutograderOutput,
    expected_status: GitAutograderStatus,
    expected_comments: List[str] = [],
) -> None:
    assert output.status == expected_status
    assert len(set(output.comments or []) & set(expected_comments)) == len(
        expected_comments
    )



from dataclasses import dataclass
from typing import Optional

from difflib_parser import DiffCode, DifflibParser
from git.diff import Diff, Lit_change_type


@dataclass
class GitAutograderDiff:
    diff: Diff
    change_type: Lit_change_type
    original_file_path: Optional[str]
    edited_file_path: Optional[str]
    original_file: Optional[str]
    edited_file: Optional[str]
    diff_parser: Optional[DifflibParser]

    def has_deleted_line(self) -> bool:
        if self.diff_parser is None:
            return False

        for diff in self.diff_parser.iter_diffs():
            if diff.code == DiffCode.LEFT_ONLY and diff.line.strip() != "":
                return True

        return False

    def has_added_line(self) -> bool:
        if self.diff_parser is None:
            return False

        for diff in self.diff_parser.iter_diffs():
            if diff.code == DiffCode.RIGHT_ONLY and diff.line.strip() != "":
                return True

        return False

    def has_edited_line(self) -> bool:
        if self.diff_parser is None:
            return False

        for diff in self.diff_parser.iter_diffs():
            if diff.code == DiffCode.CHANGED and diff.line.strip() != "":
                return True

        return False



__all__ = [
    "autograder",
    "setup_autograder",
    "set_env",
    "assert_output",
    "GitAutograderTestLoader",
    "GitAutograderRepo",
    "GitAutograderStatus",
    "GitAutograderOutput",
    "GitAutograderBranch",
    "GitAutograderRemote",
    "GitAutograderCommit",
]

from .status import GitAutograderStatus
from .output import GitAutograderOutput
from .repo import GitAutograderRepo
from .commit import GitAutograderCommit
from .branch import GitAutograderBranch
from .remote import GitAutograderRemote
from .decorators import autograder
from .test_utils import (
    setup_autograder,
    set_env,
    assert_output,
    GitAutograderTestLoader,
)
